/*
 * Softcam plugin to VDR (C++)
 *
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <vdr/sources.h>

#include "override.h"
#include "misc.h"
#include "log-core.h"

// -- cValidityRange -----------------------------------------------------------

cValidityRange::cValidityRange(void)
{
  fromCaid=toCaid=-1; fromSource=toSource=-1; fromFreq=toFreq=-1;
}

bool cValidityRange::Match(int caid, int source, int freq) const
{
  return (caid<0 || fromCaid<0 || (toCaid<0 && caid==fromCaid) || (caid>=fromCaid && caid<=toCaid)) &&
         (fromSource<0 || (toSource<0 && source==fromSource) || (source>=fromSource && source<=toSource)) &&
         (fromFreq<0 || (toFreq<0 && freq==fromFreq) || (freq>=fromFreq && freq<=toFreq));
}

char *cValidityRange::Parse3(char *s)
{
  bool log=true;
  s=skipspace(s);
  char *e=index(s,'}');
  if(e && *s++=='{') {
    *e=0; e=skipspace(e+1);
    char *p;
    if((p=index(s,':'))) {
      *p=0;
      if(ParseCaidRange(s)) {
        s=p+1;
        if((p=index(s,':'))) {
          *p=0;
          if(s==p || ParseSourceRange(s)) {
            s=p+1;
            if(!*s || ParseFreqRange(s)) return e;
            }
          log=false;
          }
        }
      else log=false;
      }
    }
  if(log) PRINTF(L_CORE_LOAD,"override: RANGE format error");
  return 0;
}

char *cValidityRange::Parse2(char *s)
{
  bool log=true;
  s=skipspace(s);
  char *e=index(s,'}');
  if(e && *s++=='{') {
    *e=0; e=skipspace(e+1);
    char *p;
    if((p=index(s,':'))) {
      *p=0;
      if(ParseSourceRange(s)) {
        s=p+1;
        if(!*s || ParseFreqRange(s)) return e;
        }
      log=false;
      }
    }
  if(log) PRINTF(L_CORE_LOAD,"override: RANGE format error");
  return 0;
}

bool cValidityRange::ParseCaidRange(const char *str)
{
  if(sscanf(str,"%x-%x",&fromCaid,&toCaid)<1) {
    PRINTF(L_CORE_LOAD,"override: CAID format error");
    return false;
    }
  if(fromCaid<0x0001 || fromCaid>0xFFFF ||
     (toCaid>0 && (toCaid<0x0001 || toCaid>0xFFFF || fromCaid>toCaid))) {
    PRINTF(L_CORE_LOAD,"override: CAID range error");
    return false;
    }
  return true;
}

bool cValidityRange::ParseFreqRange(const char *str)
{
  if(sscanf(str,"%d-%d",&fromFreq,&toFreq)<1) {
    PRINTF(L_CORE_LOAD,"override: FREQ format error");
    return false;
    }
  if(fromFreq<1 || fromFreq>50000 ||
     (toFreq>0 && (toFreq<1 || toFreq>50000 || fromFreq>toFreq))) {
    PRINTF(L_CORE_LOAD,"override: FREQ range error");
    return false;
    }
  return true;
}

bool cValidityRange::ParseSourceRange(const char *str)
{
  bool res=false;
  int l;
  char *s1=0, *s2=0;
  if((l=sscanf(str,"%a[^-:]-%a[^-:]",&s1,&s2))>=1) {
    if(s1 && (fromSource=cSource::FromString(s1))>0 &&
       (l<2 || (s2 && (toSource=cSource::FromString(s2))>0))) {
      if(cSource::IsSat(fromSource) && (toSource<0 || (cSource::IsSat(toSource) && toSource>fromSource))) {
        res=true;
        }
      else PRINTF(L_CORE_LOAD,"override: SOURCE range error");
      }
    else PRINTF(L_CORE_LOAD,"override: SOURCE parse error");
    }
  else PRINTF(L_CORE_LOAD,"override: SOURCE format error");
  free(s1); free(s2);
  return res;
}

cString cValidityRange::Print(void)
{
  char buff[256];
  int q=0;
  if(fromCaid>0) {
    q+=snprintf(buff+q,sizeof(buff)-q,"%04x",fromCaid);
    if(toCaid>0)
      q+=snprintf(buff+q,sizeof(buff)-q,"-%04x",toCaid);
    q+=snprintf(buff+q,sizeof(buff)-q,":");
    }
  if(fromSource>0) {
    q+=snprintf(buff+q,sizeof(buff)-q,"%s",*cSource::ToString(fromSource));
    if(toSource>0)
      q+=snprintf(buff+q,sizeof(buff)-q,"-%s",*cSource::ToString(toSource));
    }
  q+=snprintf(buff+q,sizeof(buff)-q,":");
  if(fromFreq>0) {
    q+=snprintf(buff+q,sizeof(buff)-q,"%d",fromFreq);
    if(toFreq>0)
      q+=snprintf(buff+q,sizeof(buff)-q,"-%d",toFreq);
    }
  return buff;
}

// -- cOverride ----------------------------------------------------------------

#define OV_CAT      1
#define OV_EMMCAID  2
#define OV_ECMTABLE 3
#define OV_EMMTABLE 4

// -- cOverrideCat -------------------------------------------------------------

class cOverrideCat : public cOverride {
private:
  int caid, pid;
public:
  cOverrideCat(void) { type=OV_CAT; }
  virtual bool Parse(char *str);
  int GetCatEntry(unsigned char *buff);
  };

bool cOverrideCat::Parse(char *str)
{
  if((str=Parse2(str))) {
    if(sscanf(str,"%x:%x",&caid,&pid)==2) {
      PRINTF(L_CORE_OVER,"cat: %s - caid %04x pid %04x",*Print(),caid,pid);
      return true;
      }
    PRINTF(L_CORE_LOAD,"override: CAT format error");
    }
  return false;
}

int cOverrideCat::GetCatEntry(unsigned char *buff)
{
  PRINTF(L_CORE_OVER,"cat: added caid %04x pid %04x",caid,pid);
  buff[0]=0x09;
  buff[1]=0x04;
  buff[2]=(caid>>8)&0xFF;
  buff[3]= caid    &0xFF;
  buff[4]=(pid >>8)&0xFF;
  buff[5]= pid     &0xFF;
  return 6;
}

// -- cOverrideEmmCaid ---------------------------------------------------------

class cOverrideEmmCaid : public cOverride {
private:
  int caid;
public:
  cOverrideEmmCaid(void) { type=OV_EMMCAID; }
  virtual bool Parse(char *str);
  int GetCaid(bool log);
  };

bool cOverrideEmmCaid::Parse(char *str)
{
  if((str=Parse3(str))) {
    if(sscanf(str,"%x",&caid)==1) {
      PRINTF(L_CORE_OVER,"emmcaid: %s - caid %04x",*Print(),caid);
      return true;
      }
    PRINTF(L_CORE_LOAD,"override: EMMCAID format error");
    }
  return false;
}

int cOverrideEmmCaid::GetCaid(bool log)
{
  if(log) PRINTF(L_CORE_OVER,"emmcaid: %04x",caid);
  return caid;
}

// -- cOverrideEcmTable --------------------------------------------------------

class cOverrideEcmTable : public cOverride {
private:
  int table;
public:
  cOverrideEcmTable(void) { type=OV_ECMTABLE; }
  virtual bool Parse(char *str);
  int GetTable(bool log);
  };

bool cOverrideEcmTable::Parse(char *str)
{
  if((str=Parse3(str))) {
    if(sscanf(str,"%x",&table)==1) {
      PRINTF(L_CORE_OVER,"ecmtable: %s - table %02x",*Print(),table);
      return true;
      }
    PRINTF(L_CORE_LOAD,"override: ECMTABLE format error");
    }
  return false;
}

int cOverrideEcmTable::GetTable(bool log)
{
  if(log) PRINTF(L_CORE_OVER,"ecmtable: %02x",table);
  return table;
}

// -- cOverrideEmmTable --------------------------------------------------------

#define OV_MAXTABLES 4

class cOverrideEmmTable : public cOverride {
public:
  int num, table[OV_MAXTABLES], mask[OV_MAXTABLES];
  //
  cOverrideEmmTable(void) { type=OV_EMMTABLE; }
  virtual bool Parse(char *str);
  void AddPids(cPids *pids, int pid, int caid);
  };

bool cOverrideEmmTable::Parse(char *str)
{
  if((str=Parse3(str))) {
    num=0;
    int n=-1;
    do {
      mask[num]=0xFF;
      int l=n+1;
      if(sscanf(&str[l],"%x%n/%x%n",&table[num],&n,&mask[num],&n)<1) {
        PRINTF(L_CORE_LOAD,"override: EMMTABLE format error");
        return false;
        }
      n+=l; num++;
      } while(num<OV_MAXTABLES && str[n]==':');
    LBSTART(L_CORE_OVER);
    LBPUT("emmtable: %s - tables",*Print());
    for(int i=0; i<num; i++) LBPUT(" %02x/%02x",table[i],mask[i]);
    LBEND();
    return true;
    }
  return false;
}

void cOverrideEmmTable::AddPids(cPids *pids, int pid, int caid)
{
  LBSTARTF(L_CORE_OVER);
  LBPUT("emmtable: %04x:",caid);
  for(int i=0; i<num; i++) {
    LBPUT(" %02x/%02x",table[i],mask[i]);
    pids->AddPid(pid,table[i],mask[i]);
    }
  LBEND();
}

// -- cOverrides ---------------------------------------------------------------

cOverrides overrides;

cOverrides::cOverrides(void)
:cStructList<cOverride>("overrides","override.conf",SL_MISSINGOK|SL_WATCH|SL_VERBOSE)
{}

cOverride *cOverrides::ParseLine(char *line)
{
  cOverride *ov=0;
  line=skipspace(line);
  char *p=index(line,':');
  if(p) {
    *p++=0;
    if(!strncasecmp(line,"cat",3)) ov=new cOverrideCat;
    else if(!strncasecmp(line,"emmcaid",7)) ov=new cOverrideEmmCaid;
    else if(!strncasecmp(line,"ecmtable",8)) ov=new cOverrideEcmTable;
    else if(!strncasecmp(line,"emmtable",8)) ov=new cOverrideEmmTable;
    if(ov && !ov->Parse(p)) { delete ov; ov=0; }
    }
  return ov;
}

cOverride *cOverrides::Find(int type, int caid, int source, int freq, cOverride *ov)
{
  for(ov=ov?Next(ov):First(); ov; ov=Next(ov))
    if(ov->Type()==type && ov->Match(caid,source,freq)) break;
  return ov;
}

int cOverrides::GetCat(int source, int transponder, unsigned char *buff, int len)
{
  int n=0;
  ListLock(false);
  for(cOverride *ov=0; (ov=Find(OV_CAT,-1,source,transponder%100000,ov)) && n<len-32;) {
    cOverrideCat *ovc=dynamic_cast<cOverrideCat *>(ov);
    if(ovc) n+=ovc->GetCatEntry(&buff[n]);
    }
  ListUnlock();
  return n;
}

void cOverrides::UpdateEcm(cEcmInfo *ecm, bool log)
{
  ListLock(false);
  cOverrideEcmTable *ovt=dynamic_cast<cOverrideEcmTable *>(Find(OV_ECMTABLE,ecm->caId,ecm->source,ecm->transponder%100000));
  if(ovt) ecm->ecm_table=ovt->GetTable(log);
  cOverrideEmmCaid *ovc=dynamic_cast<cOverrideEmmCaid *>(Find(OV_EMMCAID,ecm->caId,ecm->source,ecm->transponder%100000));
  if(ovc) ecm->emmCaId=ovc->GetCaid(log);
  ListUnlock();
}

bool cOverrides::AddEmmPids(int caid, int source, int transponder, cPids *pids, int pid)
{
  bool res=false;
  ListLock(false);
  cOverrideEmmTable *ovt=dynamic_cast<cOverrideEmmTable *>(Find(OV_EMMTABLE,caid,source,transponder%100000));
  if(ovt) {
    ovt->AddPids(pids,pid,caid);
    res=true;
    }
  ListUnlock();
  return res;
}
